/* utils.c: some useful helper functions
   Copyright (c) 1999-2004 Philip Kendall

   $Id: utils.c,v 1.40.2.2 2004/02/17 00:09:01 pak21 Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21-fuse@srcf.ucam.org
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#define _arch_dreamcast
#include <kos.h>

#include "config.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "libspectrum.h"

#include "utils.h"

extern int errno;

#define PATH_MAX	256

typedef struct path_context {

  int state;

  utils_aux_type type;
  char path[ PATH_MAX ];

} path_context;

/* Open `filename' and do something sensible with it; autoload tapes
   if `autoload' is true and return the type of file found in `type' */
int
utils_open_file( const char *filename, int autoload,
		 libspectrum_id_t *type_ptr)
{
  utils_file file;
  libspectrum_id_t type;
  libspectrum_class_t class;
  int error;

  /* Read the file into a buffer */

  if( utils_read_file( filename, &file ) ) return 1;

  /* See if we can work out what it is */
  if( libspectrum_identify_file_with_class( &type, &class, filename,
					    file.buffer, file.length ) ) {
    utils_close_file( &file );
    return 1;
  }

  error = 0;

  switch( class ) {
    
  case LIBSPECTRUM_CLASS_UNKNOWN:
    printf ( "utils_open_file: couldn't identify `%s'",
	      filename );
    utils_close_file( &file );
    return 1;

  case LIBSPECTRUM_CLASS_RECORDING:
    //error = rzx_start_playback_from_buffer( file.buffer, file.length );
    break;

  case LIBSPECTRUM_CLASS_SNAPSHOT:
    error = snapshot_read_buffer( file.buffer, file.length, type );
    break;

  case LIBSPECTRUM_CLASS_TAPE:
    error = tape_read_buffer( file.buffer, file.length, type, filename,
			      autoload );
    break;

  case LIBSPECTRUM_CLASS_DISK_PLUS3:

    printf ( "lib765 not present so can't handle .dsk files" );

    break;

  case LIBSPECTRUM_CLASS_DISK_TRDOS:
	  
	 printf ( "TRDOS not supported" );

    break;

  case LIBSPECTRUM_CLASS_CARTRIDGE_TIMEX:

	 printf ( "TIMEX not supported" );

    break;

  default:
    printf ( "utils_open_file: unknown class %d", type );
    error = 1;
    break;
  }

  if( error ) { utils_close_file( &file ); return error; }

  if( utils_close_file( &file ) ) return 1;

  if( type_ptr ) *type_ptr = type;

  return 0;
}


int
utils_read_file( const char *filename, utils_file *file )
{
  file_t fd;

  int error;

  fd = fs_open( filename, O_RDONLY);
  if( !fd ) {
    printf ( "couldn't open '%s': %d", filename,
	       errno );
    return 1;
  }

  error = utils_read_fd( fd, filename, file );
  if( error ) return error;

  return 0;
}

int
utils_read_fd( file_t fd, const char *filename, utils_file *file )
{
  stat_t file_info;
  off_t size;

  size = fs_total (fd);
  file->length = size;

  /* Either mmap() isn't available, or it failed for some reason so try
     using normal IO */

  file->mode = UTILS_FILE_OPEN_MALLOC;

  file->buffer = malloc( file->length );
  if( !file->buffer ) {
    printf ( "Out of memory at %s:%d\n", __FILE__, __LINE__ );
    return 1;
  }

  if( fs_read( fd, file->buffer, file->length ) != file->length ) 
  {
    printf ( "Error reading from '%s': %d", filename,
	      errno );
    free( file->buffer );
    fs_close( fd );
    return 1;
  }

  fs_close(fd);

  return 0;
}

int
utils_close_file( utils_file *file )
{
  switch( file->mode ) {

  case UTILS_FILE_OPEN_MMAP:

    printf ( "utils_close_file: file->mode == UTILS_FILE_OPEN_MMAP, but mmap not available?!" );

    break;

  case UTILS_FILE_OPEN_MALLOC:
    free( file->buffer );
    break;

  default:
    printf ( "Unknown file open mode %d\n", file->mode );

  }

  return 0;
}

