/*
========================================================================================
This code is used to read a .tap file, to emulate rewind, forward, start, and stop
buttons. It builds a list of levels (in a bidirectional linear list), so that the .tap
trap code can easily navigate through the levels and load them as required).
========================================================================================

Copyright (C) 2004 Stephane Schmitz

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#define _arch_dreamcast
#include <kos.h>
#include "tapreader.h"


struct stTapBlock *blockEntries = NULL;
struct stTapBlock *tapeHead = NULL;
char filename[256];
struct stTapBlock *iterator = NULL;
struct stTapBlock *hIterator = NULL;

long filePos = 0;

//#define DEBUG

int newTapReader (char *pszFilename)
{
    strcpy (filename, pszFilename);

	// Init the internal blocks list

	blockEntries= (struct stTapBlock *)malloc (sizeof(struct stTapBlock));
	blockEntries->ptrNextEntry = NULL;
	blockEntries->ptrPreviousEntry = NULL;

	filePos = 0;

    // reads levels

    if (readLevels() == NULL) 
	return (0);
	else return (1);

}

// Destroy the class instance and frees all allocated memory

void releaseTapReader (void)
{
	// Free all the linear list elements

	if (blockEntries != NULL)
	{
		struct stTapBlock *entry = getNextFreeEntry();

		// Starts from the end of the list and goes backward

		while (entry != blockEntries) // first element has been reached
		{
			entry = entry->ptrPreviousEntry;

			#ifdef DEBUG
			printf ("[DBG]: Free block (length = %d) \n", entry->dataLength);
			#endif

			//free (entry->data);
			free (entry->ptrNextEntry);
		}

		free (blockEntries);
		blockEntries = NULL;
	}
}

// Get a pointer to the next free list entry
// Sequential search ! Shouldn't be a problem though given that most ini files contain
// only some dozens keys maximum ...

struct stTapBlock *getNextFreeEntry ()
{
    struct stTapBlock *freeEntry = blockEntries;

	if (freeEntry != NULL)
	{
		while (freeEntry->ptrNextEntry != NULL)
		{
				freeEntry = freeEntry->ptrNextEntry;
		}
	}

    return (freeEntry);
}

// Read the tape file and build a list of levels

struct stTapBlock *readLevels ()
{
    file_t fd;
    unsigned char lenH, lenL, flagByte, dummy;
    int read = 0;
    long length = 0, dataBlockLength;
    struct stHeader header;
    int eof = 0, i = 0, bHasHeader = 0;
	#ifdef DEBUG
    char blockName[11];
	#endif

    fd = fs_open (filename, O_RDONLY);
    if (!fd)
    {
      return (NULL);
    }
    else
    {
      while (!eof)
      {

      // Read block length + flagbyte (00 = header, FF = block)

      read = fs_read (fd, &lenL, sizeof(char));
      if (read == 0) break;
      read = fs_read (fd, &lenH, sizeof(char));
      if (read == 0) return (NULL);
      length = lenH<<8|lenL;
      read = fs_read (fd, &flagByte, sizeof(char));
      if (read == 0) return (NULL);

	  filePos += 3;

      #ifdef DEBUG
      printf ("[DBG]: length = %d, flagbyte = %d, filePos = %d\n", length, flagByte, filePos);
      #endif

      if (flagByte == 0)
      {
        if (!addHeader (length-2, fd))
        {
          return (NULL);
        }

        // checksum

        read = fs_read (fd, &dummy, sizeof(char));
        if (read == 0) return (NULL);
		
		filePos += 1;

        bHasHeader = 1;
      }
      else
      {
        // code block
        // Add block to the dynamic list

        if (!addBlock (bHasHeader, length-2, fd)) // -1 for the checksum not read yet and -1 for the flagbyte already read
        {
          return (NULL);
        }

        // checksum

        read = fs_read (fd, &dummy, sizeof(char));
        if (read == 0) return (NULL);

		filePos += 1;

        bHasHeader = 0;

      }

      } // while !(eof)
    }

    #ifdef DEBUG
    printf ("[DBG]: fclose()\n");
    #endif

    fs_close (fd);
    return (blockEntries);
}

// Add a block entry to the blocks list

int addBlock (int bHasHeader, long length, file_t fd)
{
    long read;
	char dummy;
	int i = 0;

	struct stTapBlock *freeEntry = getNextFreeEntry();

    freeEntry->bHasHeader = bHasHeader;
	freeEntry->bIsHeader = 0;

    // Read block data
    // Allocates block data

    // freeEntry->data = (unsigned char *)malloc (length);
	// read = fs_read(fd, freeEntry->data, length);
	for (i=0; i<length; i++)
	{
		read = fs_read(fd, &dummy, 1);
		if (read == 0) return (0);
	}
	
    freeEntry->dataLength = length;
	freeEntry->dataPos = filePos;
	filePos += length;

	freeEntry->ptrNextEntry = (struct stTapBlock *)malloc (sizeof(struct stTapBlock));
	freeEntry->ptrNextEntry->ptrNextEntry = NULL;
	freeEntry->ptrNextEntry->ptrPreviousEntry = freeEntry;

    #ifdef DEBUG
    printf ("[DBG]: block added ! bHasHeader = %d\n", bHasHeader);
    #endif

    return (1);
}

// Add a block entry to the blocks list

int addHeader (long length, file_t fd)
{
    long read;
	long dataBlockLength;
	struct stHeader header;
	#ifdef DEBUG
    char blockName[11];
	#endif

	// header

    read = fs_read (fd, &header, sizeof(struct stHeader));
    if (read == 0) return (0);

    #ifdef DEBUG
	strncpy (blockName, header.filename, 10);
    blockName[10] = '\0';
    printf ("[DBG]: blockName = %s\n", blockName);
    #endif

    // length of data block

    dataBlockLength = header.dataBlockLenH<<8|header.dataBlockLenL;

    #ifdef DEBUG
    printf ("[DBG]: header block type = %d, dataBlockLength = %d\n", header.blockType, dataBlockLength);
    #endif

	struct stTapBlock *freeEntry = getNextFreeEntry();

    freeEntry->bHasHeader = 1;
	freeEntry->bIsHeader = 1;
    freeEntry->header = header;

    // Read block data
    // Allocates block data

    //freeEntry->data = (unsigned char *)malloc (length);
	//memcpy (freeEntry->data, &header, sizeof (struct stHeader));
	freeEntry->dataPos = filePos;
	filePos += sizeof(struct stHeader);
    freeEntry->dataLength = length;
	freeEntry->ptrNextEntry = (struct stTapBlock *)malloc (sizeof(struct stTapBlock));
	freeEntry->ptrNextEntry->ptrNextEntry = NULL;
	freeEntry->ptrNextEntry->ptrPreviousEntry = freeEntry;

    #ifdef DEBUG
    printf ("[DBG]: header added ! \n");
    #endif

	return (1);
}

// Get the next block

struct stTapBlock *getNextBlock ()
{
    int bFound = 0;

    if (tapeHead == NULL)
    {
		tapeHead = blockEntries;
    }
	else
	{
		if (tapeHead->ptrNextEntry != NULL)
		{
			tapeHead = tapeHead->ptrNextEntry;
		}
	}

    return (tapeHead);
}

// Get the previous block

struct stTapBlock *getPreviousBlock ()
{
    int bFound = 0;

    if (tapeHead == NULL)
    {
		tapeHead = blockEntries;
    }
	else
	{
		if (tapeHead != blockEntries) // first element has been reached
		{
			tapeHead = tapeHead->ptrPreviousEntry;
		}

	}

    return (tapeHead);
}

// Looks for a block and returns a pointer to this block

struct stTapBlock *searchBlock (long length, unsigned char blockType)
{
    struct stTapBlock *entry = blockEntries;
    int bFound = 0;

    while (entry->ptrNextEntry != NULL && !bFound)
    {
        if (entry->dataLength == length && (entry->bHasHeader && entry->header.blockType == blockType) )
        {
                bFound = 1;
        }
        else
        {
                entry = entry->ptrNextEntry;
        }
    }

    if (!bFound) return (NULL); else return (entry);
}

// Looks for a block and returns a pointer to this block

struct stTapBlock *searchBlockByLength (long length)
{
    struct stTapBlock *entry = blockEntries;
    int bFound = 0;

    while (entry->ptrNextEntry != NULL && !bFound)
    {
		#ifdef DEBUG
		printf ("searchBlockByLength() : length = %d, entry->dataLength = %d\n", length, entry->dataLength);
		#endif

        if (entry->dataLength == length)
        {
                bFound = 1;
        }
        else
        {
                entry = entry->ptrNextEntry;
        }
    }

    if (!bFound) return (NULL); else return (entry);
}

// Blocks iterator init

char *getBlockDescription (struct stTapBlock *block)
{
    if (block->bIsHeader)
    {
      switch (block->header.blockType)
      {
        case 0:
          return ("Program");
        case 1:
          return ("Number array");
        case 2:
          return ("Character array");
        case 3:
          if (block->dataLength==6912)
          {
            return ("Code (Screen$)");
          }
          else
          {
            return ("Code");
          }
      }
    }
    else
    {
      return ("Unknown");
    }
}

// Blocks iterator init

struct stTapBlock *blockIteratorInit ()
{
    iterator = blockEntries;
    return (iterator);
}

// Get next entry from block iterator

struct stTapBlock *blockIteratorNext ()
{
    if (iterator->ptrNextEntry->ptrNextEntry != NULL)
    {
      iterator = iterator->ptrNextEntry;
      return (iterator);
    }
    else
    {
      return (NULL);
    }
}

// Headers iterator init

struct stTapBlock *headerIteratorInit ()
{
    hIterator = blockEntries;
    return (hIterator);
}

// Get next header from headers iterator

struct stTapBlock *headerIteratorNext ()
{
    if (hIterator->ptrNextEntry->ptrNextEntry != NULL)
    {
      hIterator = hIterator->ptrNextEntry;
	  if (!hIterator->bIsHeader)
	  {
			if (hIterator->ptrNextEntry->ptrNextEntry != NULL)
			{
				hIterator = hIterator->ptrNextEntry;
			}
	  }

      return (hIterator);
    }
    else
    {
      return (NULL);
    }
}

int tape_open (char *filename)
{
	int rc;

	rc = newTapReader(filename);
   
	struct stTapBlock *block;
	char blockName[11];

	#ifdef DEBUG
	for (block = blockIteratorInit(); block;
           block = blockIteratorNext())
	{
		strncpy(blockName, block -> bHasHeader ? block -> header.filename : "(UNKNOWN)", 10);
		blockName[10] = '\0';
		printf("Iterator: name = %s, size = %d, type = %s, dataPos = %d\n", blockName, block -> dataLength, getBlockDescription(block), block->dataPos);
	}
	#endif

	return (rc);
}

int tape_close ()
{
   releaseTapReader();
   tapeHead = NULL;
   return (1);
}

int tape_rewind ()
{
  tapeHead = NULL;
  return (1);
}

int tape_play ()
{
  return (1);
}

int tape_stop ()
{
   return (1);
}
